use std::{
	cmp::min,
	env,
	sync::{Arc, Mutex},
};

use easy_from::EasyFrom;
use rusqlite::{Connection, Row};
use serde::Serialize;
use serenity::{
	async_trait,
	model::prelude::{
		command::{Command, CommandOptionType, CommandType},
		component::{ActionRowComponent, InputTextStyle},
		interaction::{
			application_command::ApplicationCommandInteraction,
			autocomplete::AutocompleteInteraction,
			modal::ModalSubmitInteraction, Interaction,
			InteractionResponseType,
		},
		Ready,
	},
	prelude::{Context, EventHandler, GatewayIntents},
	Client,
};

use crate::db_types::{heroes, responses};

type SharedConnection = Arc<Mutex<Connection>>;


#[derive(Debug)]
enum NameOrId {
	Name(String),
	Id(usize),
}
#[derive(Debug)]
struct ResponseCommandInput {
	hero:           Option<NameOrId>,
	response_id:    Option<usize>,
	processed_text: String,
}
impl ResponseCommandInput {
	fn parse(
		input: &str,
	) -> core::result::Result<Self, core::num::ParseIntError> {
		let mut hero = None;
		let mut response_id = None;
		let mut pieces = Vec::new();

		let words = input.split_whitespace();

		for word in words {
			if let Some(name_or_id_str) = word.strip_prefix("@hero:") {
				match name_or_id_str.parse::<usize>() {
					Ok(u) => hero = Some(NameOrId::Id(u)),
					Err(_) => {
						hero = Some(NameOrId::Name(
							name_or_id_str
								.replace(|c: char| !c.is_alphanumeric(), ""),
						))
					}
				}

				continue;
			}

			if let Some(id) = word.strip_prefix("@response_id:") {
				let id: usize = id.parse()?;
				response_id = Some(id);

				continue;
			}

			let word = word.replace(|c: char| !c.is_alphanumeric(), "");
			pieces.push(word);
		}

		let processed_text = pieces.join(" ");

		Ok(Self {
			hero,
			response_id,
			processed_text,
		})
	}
}

#[derive(Clone)]
struct HeroResponse {
	text:      responses::OriginalTextOwned,
	audio_url: responses::AudioUrlOwned,
	hero_name: heroes::NameOwned,
	hero_icon: heroes::IconUrlOwned,
}

fn choose_response(
	conn: &SharedConnection,
	input: &str,
) -> Result<Option<HeroResponse>> {
	let input = ResponseCommandInput::parse(input)?;

	let results: Vec<HeroResponse> = {
		let conn = conn.lock().unwrap();
		let map_row = |r: &Row| -> rusqlite::Result<HeroResponse> {
			let hero_name = heroes::NameOwned(r.get(0)?);
			let icon_url = heroes::IconUrlOwned(r.get(1)?);
			let original_text = responses::OriginalTextOwned(r.get(2)?);
			let audio_url = responses::AudioUrlOwned(r.get(3)?);

			Ok(HeroResponse {
				hero_name,
				hero_icon: icon_url,
				text: original_text,
				audio_url,
			})
		};

		let map_err =
			|r: rusqlite::Result<HeroResponse>| -> Result<HeroResponse> {
				r.map_err(Error::from)
			};

		match (&input.response_id, &input.hero) {
			(Some(id), _) => {
				let mut stmt = conn.prepare(
					r"
					SELECT name, icon_url, original_text, audio_url
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE r.id = ?
					",
				)?;

				let results = stmt
					.query_map((id,), map_row)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
			(_, None) => {
				let mut stmt = conn.prepare(
					r"
					SELECT name, icon_url, original_text, audio_url
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE processed_text LIKE ?
					",
				)?;

				let results = stmt
					.query_map((input.processed_text.as_str(),), map_row)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
			(_, Some(NameOrId::Id(hero_id))) => {
				let mut stmt = conn.prepare(
					r"
					SELECT name, icon_url, original_text, audio_url
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE
						processed_text LIKE ?
						AND h.id = ?
					",
				)?;

				let results = stmt
					.query_map(
						(input.processed_text.as_str(), hero_id),
						map_row,
					)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
			(_, Some(NameOrId::Name(hero_name))) => {
				let hero_name = format!("%{}%", hero_name);

				let mut stmt = conn.prepare(
					r"
					SELECT name, icon_url, original_text, audio_url
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE
						processed_text LIKE ?
						AND h.name LIKE ?
					",
				)?;

				let results = stmt
					.query_map(
						(input.processed_text.as_str(), hero_name),
						map_row,
					)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
		}
	};

	if results.is_empty() {
		return Ok(None);
	}

	use rand::seq::SliceRandom;

	Ok(results.choose(&mut rand::thread_rng()).cloned())
}

async fn handle_response_command(
	conn: &SharedConnection,
	cx: Context,
	ac: ApplicationCommandInteraction,
) -> Result<()> {
	let input = match ac.data.options.iter().find(|o| o.name == "voice_line") {
		Some(input) => input,
		None => return Ok(()),
	};
	let input = input.value.as_ref().unwrap();
	let input = input.as_str().unwrap();

	let response = match choose_response(conn, input) {
		Ok(Some(res)) => res,
		Ok(None) => {
			ac.create_interaction_response(&cx.http, |ir| {
				ir.kind(InteractionResponseType::ChannelMessageWithSource)
					.interaction_response_data(|rd| {
						rd.content("No responses matched").ephemeral(true)
					})
			})
			.await?;

			return Ok(());
		}
		Err(Error::ParseIntError(_)) => {
			ac.create_interaction_response(&cx.http, |ir| {
				ir.kind(InteractionResponseType::ChannelMessageWithSource)
					.interaction_response_data(|rd| {
						rd.content("Invalid input").ephemeral(true)
					})
			})
			.await?;

			return Ok(());
		}
		Err(e) => {
			return Err(e);
		}
	};

	ac.create_interaction_response(&cx.http, |ir| {
		ir.kind(InteractionResponseType::ChannelMessageWithSource)
			.interaction_response_data(|rd| {
				rd.embed(|e| {
					e.author(|a| {
						a.name(&response.hero_name.0)
							.icon_url(&response.hero_icon.0)
							.url(format!(
								"https://dota2.fandom.com/wiki/{}",
								response.hero_name.0.replace(' ', "_"),
							))
					})
					.description(format!(
						"[{}]({})",
						response.text.0, response.audio_url.0
					))
				})
			})
	})
	.await?;

	Ok(())
}

async fn handle_reply_with_response_command(
	cx: Context,
	ac: ApplicationCommandInteraction,
) -> Result<()> {
	let message_id = ac.data.resolved.messages.values().next().unwrap().id;

	ac.create_interaction_response(&cx.http, |ir| {
		ir.kind(InteractionResponseType::Modal)
			.interaction_response_data(|ird| {
				ird.title("Reply with response")
					.components(|c| {
						c.create_action_row(|r| {
							r.create_input_text(|it| {
								it.placeholder("Shitty Wizard!")
									.label("Voice line")
									.style(InputTextStyle::Short)
									.custom_id("voice_line")
									.required(true)
							})
						})
					})
					.custom_id(format!("reply_with_response:{}", message_id))
			})
	})
	.await?;

	Ok(())
}

async fn handle_application_command(
	conn: &SharedConnection,
	cx: Context,
	ac: ApplicationCommandInteraction,
) -> Result<()> {
	match ac.data.name.as_str() {
		"response" => handle_response_command(conn, cx, ac).await,
		"Reply with response" => {
			handle_reply_with_response_command(cx, ac).await
		}
		_ => {
			ac.create_interaction_response(&cx.http, |ir| {
				ir.kind(InteractionResponseType::ChannelMessageWithSource)
					.interaction_response_data(|ird| {
						ird.content("Unknown command").ephemeral(true)
					})
			})
			.await?;

			Ok(())
		}
	}
}

async fn handle_response_autocomplete(
	conn: &SharedConnection,
	cx: Context,
	ac: AutocompleteInteraction,
) -> Result<()> {
	let input = match ac.data.options.iter().find(|o| o.name == "voice_line") {
		Some(input) => input,
		None => return Ok(()),
	};

	let input = input.value.as_ref().unwrap();
	let input = input.as_str().unwrap();
	let input = match ResponseCommandInput::parse(input) {
		Ok(i) => i,
		Err(_) => {
			ac.create_autocomplete_response(&cx.http, |r| {
				r.set_choices(serde_json::Value::Array(Vec::new()))
			})
			.await?;
			return Ok(());
		}
	};

	let fuzzy_input = format!("%{}%", input.processed_text.replace(' ', "%"));

	let suggestions: Vec<Suggestion> = {
		let map_row = |r: &Row| -> rusqlite::Result<Suggestion> {
			let response_id = responses::Id(r.get(0)?);
			let original_text = responses::OriginalTextOwned(r.get(1)?);
			let hero_name = heroes::NameOwned(r.get(2)?);

			let name = format!("{}: {}", hero_name.0, original_text.0);
			let value = format!("@response_id:{}", response_id.0);

			let name = name[0..min(100, name.len())].to_string();
			let value = value[0..min(100, value.len())].to_string();

			Ok(Suggestion { name, value })
		};
		let map_err = |r: rusqlite::Result<Suggestion>| -> Result<Suggestion> {
			r.map_err(Error::from)
		};

		let conn = conn.lock().unwrap();
		match (&input.response_id, &input.hero) {
			(Some(id), _) => {
				let mut stmt = conn.prepare(
					r"
					SELECT r.id, original_text, h.name
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE r.id = ?
					",
				)?;

				let results = stmt
					.query_map((id,), map_row)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
			(_, None) => {
				let mut stmt = conn.prepare(
					r"
					SELECT r.id, original_text, h.name
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE r.processed_text LIKE ?
					ORDER BY RANDOM()
					LIMIT 25
					",
				)?;

				let results = stmt
					.query_map((fuzzy_input.as_str(),), map_row)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
			(_, Some(NameOrId::Id(hero_id))) => {
				let mut stmt = conn.prepare(
					r"
					SELECT r.id, original_text, h.name
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE
						r.processed_text LIKE ?
						AND h.id = ?
					ORDER BY RANDOM()
					LIMIT 25
					",
				)?;

				let results = stmt
					.query_map((fuzzy_input.as_str(), hero_id), map_row)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
			(_, Some(NameOrId::Name(hero_name))) => {
				let hero_name = format!("%{}%", hero_name);
				let mut stmt = conn.prepare(
					r"
					SELECT r.id, original_text, h.name
					FROM responses AS r
					INNER JOIN heroes AS h ON h.id = r.hero_id
					WHERE
						r.processed_text LIKE ?
						AND h.name LIKE ?
					ORDER BY RANDOM()
					LIMIT 25
					",
				)?;

				let results = stmt
					.query_map((fuzzy_input.as_str(), hero_name), map_row)?
					.map(map_err)
					.collect::<Result<_>>()?;

				results
			}
		}
	};

	let suggestions = serde_json::to_value(suggestions)?;

	ac.create_autocomplete_response(cx.http, |r| r.set_choices(suggestions))
		.await?;

	Ok(())
}

async fn handle_autocomplete(
	conn: &SharedConnection,
	cx: Context,
	ac: AutocompleteInteraction,
) -> Result<()> {
	match ac.data.name.as_str() {
		"response" => handle_response_autocomplete(conn, cx, ac).await,
		_ => Ok(()),
	}
}

async fn handle_reply_with_response_modal(
	conn: &SharedConnection,
	cx: Context,
	ms: ModalSubmitInteraction,
	message_id: String,
) -> Result<()> {
	ms.create_interaction_response(&cx.http, |ir| {
		ir.kind(InteractionResponseType::DeferredChannelMessageWithSource)
			.interaction_response_data(|ird| ird.ephemeral(true))
	})
	.await?;

	let mut components =
		ms.data.components.iter().flat_map(|r| r.components.iter());
	let message_id = message_id.parse::<u64>().unwrap();

	let voice_line = components
		.find_map(|c| match c {
			ActionRowComponent::InputText(it) => Some(&it.value),
			_ => None,
		})
		.unwrap();

	let message = ms.channel_id.message(&cx.http, message_id).await?;

	let response = match choose_response(conn, voice_line) {
		Ok(None) => {
			ms.edit_original_interaction_response(&cx.http, |ir| {
				ir.content("No responses matched")
			})
			.await?;

			return Ok(());
		}
		Ok(Some(response)) => response,
		Err(Error::ParseIntError(_)) => {
			ms.edit_original_interaction_response(&cx.http, |ir| {
				ir.content("Invalid input")
			})
			.await?;

			return Ok(());
		}
		Err(e) => return Err(e),
	};

	let user_string = match ms.guild_id {
		None => format!(
			"@{}#{} used Reply with response",
			ms.user.name, ms.user.discriminator
		),
		Some(gi) => match ms.user.nick_in(&cx.http, gi).await {
			Some(nick) => format!("@{} used Reply with response", nick),
			None => format!(
				"@{}#{} used Reply with response",
				ms.user.name, ms.user.discriminator
			),
		},
	};
	ms.channel_id
		.send_message(&cx.http, |m| {
			m.embed(|e| {
				e.author(|a| {
					a.name(&response.hero_name.0)
						.icon_url(&response.hero_icon.0)
						.url(format!(
							"https://dota2.fandom.com/wiki/{}",
							response.hero_name.0.replace(' ', "_"),
						))
				})
				.description(format!(
					"[{}]({})",
					response.text.0, response.audio_url.0,
				))
				.footer(|f| {
					match &ms.user.avatar_url() {
						Some(u) => f.icon_url(u),
						None => f,
					}
					.text(user_string)
				})
			})
			.reference_message(&message)
		})
		.await?;

	ms.edit_original_interaction_response(&cx.http, |ir| {
		ir.content("Reply sent")
	})
	.await?;

	Ok(())
}

async fn handle_modal_submit(
	conn: &SharedConnection,
	cx: Context,
	ms: ModalSubmitInteraction,
) -> Result<()> {
	if let Some(message_id) = &ms
		.data
		.custom_id
		.as_str()
		.strip_prefix("reply_with_response:")
	{
		let message_id = message_id.to_string();
		handle_reply_with_response_modal(conn, cx, ms, message_id).await?;

		return Ok(());
	}

	eprintln!("Unknown modal id {}", ms.data.custom_id);

	Ok(())
}

#[derive(Serialize, Debug)]
struct Suggestion {
	name:  String,
	value: String,
}

struct Handler {
	conn: SharedConnection,
}
#[async_trait]
impl EventHandler for Handler {
	async fn interaction_create(&self, cx: Context, interaction: Interaction) {
		let res: Result<()> = async {
			match interaction {
				Interaction::ApplicationCommand(ac) => {
					handle_application_command(&self.conn, cx, ac).await
				}
				Interaction::Autocomplete(ac) => {
					handle_autocomplete(&self.conn, cx, ac).await
				}
				Interaction::ModalSubmit(ms) => {
					handle_modal_submit(&self.conn, cx, ms).await
				}
				_ => Ok(()),
			}
		}
		.await;

		if let Err(e) = res {
			eprintln!("{:?}", e);
		}
	}

	async fn ready(&self, cx: Context, ready: Ready) {
		println!("{} connected", ready.user.name);

		match Command::create_global_application_command(&cx.http, |c| {
			c.name("response")
				.description("Send a response voice line")
				.kind(CommandType::ChatInput)
				.create_option(|c| {
					c.kind(CommandOptionType::String)
						.name("voice_line")
						.description("the voice line to send")
						.set_autocomplete(true)
						.required(true)
				})
		})
		.await
		{
			Ok(_) => println!("Command registered"),
			Err(e) => eprintln!("Failed to register command: {:?}", e),
		};

		match Command::create_global_application_command(&cx.http, |c| {
			c.name("Reply with response").kind(CommandType::Message)
		})
		.await
		{
			Ok(_) => println!("Command registered"),
			Err(e) => eprintln!("Failed to register command: {:?}", e),
		};
	}
}


#[derive(EasyFrom, Debug)]
pub enum Error {
	Serenity(#[from] serenity::Error),
	Rusqlite(#[from] rusqlite::Error),
	SerdeJson(#[from] serde_json::Error),
	ParseIntError(#[from] core::num::ParseIntError),
}

pub type Result<T> = core::result::Result<T, Error>;

pub async fn run_bot(conn: Connection) -> Result<()> {
	let token = env::var("DISCORD_TOKEN").expect("missing DISCORD_TOKEN");
	let intents = GatewayIntents::empty();
	let mut client = Client::builder(&token, intents)
		.event_handler(Handler {
			conn: Arc::new(Mutex::new(conn)),
		})
		.await?;


	client.start().await?;

	Ok(())
}
