use std::{
	collections::{HashMap, VecDeque},
	sync::{
		atomic::{AtomicU64, Ordering},
		Arc, Mutex,
	},
};

use easy_from::EasyFrom;
use futures::{
	future::{join_all, try_join_all},
	try_join,
};
use once_cell::sync::Lazy;
use regex::Regex;
use rusqlite::Connection;
use serde::Deserialize;
use tokio::sync::Barrier;

use crate::db_types::{heroes, responses};

fn check_status(res: reqwest::Response) -> Result<reqwest::Response> {
	if !res.status().is_success() {
		return Err(Error::HttpUnsuccessful(res.status()));
	}

	Ok(res)
}

async fn parse_response_json<R>(res: reqwest::Response) -> Result<R>
where
	R: for<'a> Deserialize<'a>,
{
	let res = check_status(res)?;

	Ok(res.json::<R>().await?)
}

#[derive(Debug, EasyFrom)]
pub enum Error {
	Rusqlite(#[from] rusqlite::Error),
	Reqwest(#[from] reqwest::Error),
	HttpUnsuccessful(reqwest::StatusCode),
	NoIconRegexMatch,
}
pub type Result<T> = core::result::Result<T, Error>;

pub fn initialize_db(conn: &mut Connection) -> Result<()> {
	conn.execute_batch(
		r"
		DROP TABLE IF EXISTS responses;
		DROP TABLE IF EXISTS heroes;

		CREATE TABLE heroes (
			id INTEGER PRIMARY KEY,
			name TEXT NOT NULL,
			icon_url TEXT NOT NULL
		) STRICT;

		CREATE TABLE responses (
			id INTEGER PRIMARY KEY,
			hero_id INTEGER NOT NULL,
			processed_text TEXT NOT NULL,
			original_text TEXT NOT NULL,
			audio_url TEXT NOT NULL,
			FOREIGN KEY(hero_id) REFERENCES heroes(id)
		) STRICT;
		",
	)?;

	Ok(())
}

struct RequestLimiter {
	queue:         Arc<Mutex<VecDeque<Arc<Barrier>>>>,
	in_flight:     AtomicU64,
	max_in_flight: usize,
}
impl RequestLimiter {
	fn new(max_in_flight: usize) -> Self {
		RequestLimiter {
			queue: Arc::new(Mutex::new(VecDeque::new())),
			in_flight: AtomicU64::new(0),
			max_in_flight,
		}
	}

	async fn unblock_one(&self) {
		let next = match self.queue.lock().unwrap().pop_front() {
			Some(b) => b,
			None => return,
		};

		next.wait().await;
	}

	async fn get(&self, url: String) -> reqwest::Result<reqwest::Response> {
		loop {
			let in_flight = self.in_flight.load(Ordering::Relaxed);
			let next_in_flight = in_flight + 1;
			if next_in_flight < self.max_in_flight as u64 {
				if self
					.in_flight
					.compare_exchange(
						in_flight,
						next_in_flight,
						Ordering::Relaxed,
						Ordering::Relaxed,
					)
					.is_err()
				{
					continue;
				}

				let res = reqwest::get(url).await;
				self.in_flight.fetch_sub(1, Ordering::Relaxed);
				self.unblock_one().await;

				return res;
			}

			let barrier = Arc::new(Barrier::new(2));
			self.queue.lock().unwrap().push_back(barrier.clone());

			barrier.wait().await;
		}
	}
}

pub async fn fetch_and_insert_data(conn: &mut Connection) -> Result<()> {
	let limiter = RequestLimiter::new(128);

	let hero_pages = fetch_response_page_urls(&limiter).await?;
	let futures = hero_pages
		.into_iter()
		.map(|p| get_heroes(p, &limiter))
		.collect::<Vec<_>>();

	let heroes =
		join_all(futures)
			.await
			.into_iter()
			.flat_map(|(hn, r)| match r {
				Err(_) => {
					eprintln!("Failed to fetch {}: {:?}", hn, r);

					None
				}
				Ok(r) => Some(r),
			});

	println!("All fetched, inserting into DB");
	for h in heroes {
		let tx = conn.transaction()?;
		{
			let mut insert_response = tx.prepare(
				"INSERT INTO responses (hero_id, processed_text, \
				 original_text, audio_url) VALUES (?, ?, ?, ?)",
			)?;

			let mut insert_hero_icon = tx.prepare(
				"INSERT INTO heroes (name, icon_url) VALUES (?, ?) RETURNING \
				 id",
			)?;


			let hero_id: heroes::Id = insert_hero_icon
				.query_row((h.name.0, h.icon_url.0), |r| {
					Ok(heroes::Id(r.get(0)?))
				})?;


			for response in h.responses {
				insert_response.execute((
					hero_id.0,
					response.processed_text.0,
					response.original_text.0,
					response.audio_url.0,
				))?;
			}
		}

		tx.commit()?;
	}

	Ok(())
}

#[derive(Deserialize, Debug)]
struct CategoryMember {
	title: String,
}
#[derive(Deserialize, Debug)]
struct CategoryMembers {
	categorymembers: Vec<CategoryMember>,
}
#[derive(Deserialize, Debug)]
struct HeroesResponse {
	query: CategoryMembers,
}

#[derive(Debug)]
struct HeroUrls {
	hero_name:     String,
	responses_url: String,
	hero_page_url: String,
}
const API_BASE: &str = "https://dota2.fandom.com";
async fn fetch_response_page_urls(
	limiter: &RequestLimiter,
) -> Result<Vec<HeroUrls>> {
	let hero_endpoint = format!(
		"{}/api.php?action=query&list=categorymembers&cmtitle=Category:Heroes&\
		 cmlimit=max&cmtype=page&format=json",
		API_BASE
	);
	let resp = limiter.get(hero_endpoint).await?;
	let resp: HeroesResponse = parse_response_json(resp).await?;

	Ok(resp
		.query
		.categorymembers
		.into_iter()
		.map(|cm| HeroUrls {
			hero_name:     cm.title.clone(),
			responses_url: format!(
				"{}/wiki/{}/Responses?action=raw",
				API_BASE, cm.title
			),
			hero_page_url: format!("{}/wiki/{}?action=raw", API_BASE, cm.title),
		})
		.collect())
}

#[derive(Debug)]
struct Response {
	processed_text: responses::ProcessedTextOwned,
	original_text:  responses::OriginalTextOwned,
	audio_url:      responses::AudioUrlOwned,
}

#[derive(Debug)]
struct Hero {
	name:      heroes::NameOwned,
	icon_url:  heroes::IconUrlOwned,
	responses: Vec<Response>,
}

async fn get_heroes(
	urls: HeroUrls,
	limiter: &RequestLimiter,
) -> (String, Result<Hero>) {
	let res = (|| async {
		static RESPONSE: Lazy<Regex> = Lazy::new(|| {
			Regex::new(
				r"(?m)(?:^\*\s+<sm2>(?P<file>.*?)</sm2>(?:.*<sm2>.*</sm2>)*)(?:.*\{\{.*?\}\}.*?)* ?(?P<text>.*$)",
			)
			.unwrap()
		});

		println!("Fetching {}", urls.hero_name);

		let get_responses = (|| async {
			let resp = limiter.get(urls.responses_url).await?;
			if !resp.status().is_success() {
				panic!(
					"Failed to fetch hero responses: {} {}",
					urls.hero_name,
					resp.status()
				);
			}
			let body = resp.text().await?;

			let responses: Vec<_> = RESPONSE
				.captures_iter(body.as_str())
				.map(|c| {
					let file = c.name("file").unwrap();
					let response = c.name("text").unwrap();

					Ok((response.as_str(), OriginalFilename(file.as_str())))
				})
				.collect::<Result<_>>()?;

			let filenames: Vec<_> =
				responses.iter().map(|(_, file)| *file).collect();
			let mut filenames_to_urls =
				fetch_all_file_urls(&filenames, limiter).await?;
			let responses = responses
				.into_iter()
				.flat_map(|(response, filename)| {
					let url = match filenames_to_urls.remove(filename.0) {
						None => return None,
						Some(url) => url,
					};

					Some(Response {
						processed_text: process_response(response),
						original_text:  responses::OriginalTextOwned(
							response.replace(|c| {
								c == '[' || c == ']'
							}, ""),
						),
						audio_url:      responses::AudioUrlOwned(url.0),
					})
				})
				.collect();

			println!("Fetched responses for {}", urls.hero_name);

			Result::Ok(responses)
		})();

		let get_icon_url = (|| async {
			let res = limiter.get(urls.hero_page_url).await?;
			let res = check_status(res)?;
			let body = res.text().await?;

			static ICON: Lazy<Regex> = Lazy::new(|| {
				Regex::new(r"(?m)^\s*\|\s*icon\s*=\s*(?P<icon>.*)$").unwrap()
			});

			let icon_filename = ICON
				.captures_iter(body.as_str())
				.next()
				.ok_or(Error::NoIconRegexMatch)?
				.name("icon")
				.unwrap()
				.as_str();
			let res = fetch_hero_icon_url(icon_filename, limiter).await?;

			println!("Got icon for {}", urls.hero_name);

			Result::Ok(res)
		})();

		let (responses, icon_url) = try_join!(get_responses, get_icon_url)?;

		Ok(Hero {
			name: heroes::NameOwned(urls.hero_name.clone()),
			icon_url: heroes::IconUrlOwned(icon_url),
			responses,
		})
	})()
	.await;

	(urls.hero_name, res)
}

fn process_response(s: &str) -> responses::ProcessedTextOwned {
	responses::ProcessedTextOwned(
		s.to_lowercase()
			.replace(|c: char| !c.is_whitespace() && !c.is_alphanumeric(), ""),
	)
}

#[derive(Debug, Deserialize)]
struct ImageInfo {
	url: String,
}
#[derive(Debug, Deserialize)]
struct ImageInfoPage {
	imageinfo: Option<Vec<ImageInfo>>,
	title:     String,
}

#[derive(Debug, Deserialize)]
struct NormalizedFilename {
	from: String,
	to:   String,
}

#[derive(Debug, Deserialize)]
struct ImageInfoQueryWithNormalized {
	normalized: Vec<NormalizedFilename>,
	pages:      HashMap<String, ImageInfoPage>,
}

#[derive(Debug, Deserialize)]
struct ImageInfoQuery {
	pages: HashMap<String, ImageInfoPage>,
}

#[derive(Debug, Deserialize)]
struct ImageInfoResponseBase<Body> {
	query: Body,
}

type ImageInfoWithNormalizedResponse =
	ImageInfoResponseBase<ImageInfoQueryWithNormalized>;
type ImageInfoResponse = ImageInfoResponseBase<ImageInfoQuery>;

#[derive(Clone, Hash, Eq, PartialEq)]
struct OriginalFilenameOwned(String);

#[derive(Clone, Hash, Eq, PartialEq, Copy)]
struct OriginalFilename<'a>(&'a str);
impl core::borrow::Borrow<str> for OriginalFilenameOwned {
	fn borrow(&self) -> &str {
		self.0.as_str()
	}
}

#[derive(Clone)]
struct FileUrlOwned(String);

async fn fetch_all_file_urls(
	filenames: &[OriginalFilename<'_>],
	limiter: &RequestLimiter,
) -> Result<HashMap<OriginalFilenameOwned, FileUrlOwned>> {
	let futures = filenames
		.chunks(50)
		.map(|filenames| async {
			let fnames = filenames
				.iter()
				.map(|fname| format!("File:{}", fname.0))
				.collect::<Vec<_>>()
				.join("|");
			let url = format!(
				"{}/api.php?action=query&prop=imageinfo&iiprop=url&\
				 format=json&titles={}",
				API_BASE, fnames,
			);

			let resp = limiter.get(url).await?;
			let res: ImageInfoWithNormalizedResponse =
				parse_response_json(resp).await?;

			let normalized_to_original: HashMap<&str, OriginalFilenameOwned> =
				res.query
					.normalized
					.iter()
					.map(|n| {
						(
							n.to.as_str(),
							OriginalFilenameOwned(
								n.from.as_str().replace("File:", ""),
							),
						)
					})
					.collect();

			let original_to_url: HashMap<OriginalFilenameOwned, FileUrlOwned> =
				res.query
					.pages
					.into_values()
					.flat_map(|p| {
						let title = p.title;
						let url = match p.imageinfo {
							Some(ii) => ii.into_iter().next().unwrap().url,
							None => return None,
						};

						Some((
							normalized_to_original
								.get(title.as_str())
								.unwrap()
								.clone(),
							FileUrlOwned(url),
						))
					})
					.collect();

			Result::Ok(original_to_url)
		})
		.collect::<Vec<_>>();

	let maps = try_join_all(futures).await?;
	let map = maps.into_iter().fold(HashMap::new(), |mut acc, map| {
		acc.extend(map.into_iter());

		acc
	});

	Ok(map)
}

async fn fetch_hero_icon_url(
	filename: &str,
	limiter: &RequestLimiter,
) -> Result<String> {
	let url = format!(
		"{}/api.php?action=query&prop=imageinfo&iiprop=url&format=json&\
		 titles={}",
		API_BASE, filename
	);
	let res = limiter.get(url).await?;
	let res: ImageInfoResponse = parse_response_json(res).await?;

	let url = res
		.query
		.pages
		.into_values()
		.next()
		.unwrap()
		.imageinfo
		.unwrap()
		.into_iter()
		.next()
		.unwrap()
		.url;

	Ok(url)
}
